/*
 * Copyright (c) 1999 by Sun Microsystems, Inc.
 * All rights reserved.
 */

/*
 * Portions may be based on code from the linux ext2 filesystem
 *
 *  Copyright (C) 1993  Stephen Tweedie (sct@dcs.ed.ac.uk)
 *  Copyright (C) 1992  Remy Card (card@masi.ibp.fr)
 *                      Laboratoire MASI - Institut Blaise Pascal
 *                      Universite Pierre et Marie Curie (Paris VI)
 * Copyright (C) 1991, 1992  Linus Torvalds
 */

#pragma ident	"%Z%%M%	%I%	%E% SMI"

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kmem.h>
#include <sys/user.h>
#include <sys/proc.h>
#include <sys/cred.h>
#include <sys/disp.h>
#include <sys/buf.h>
#include <sys/vfs.h>
#include <sys/vnode.h>
#include <sys/fdio.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/conf.h>
#undef NFSCLIENT
#include <sys/statvfs.h>
#include <sys/mount.h>
#include <sys/pathname.h>
#include <sys/cmn_err.h>
#include <sys/debug.h>
#include <sys/sysmacros.h>
#include <sys/conf.h>
#include <sys/mkdev.h>
#include <sys/swap.h>
#include <sys/sunddi.h>
#include <sys/dktp/fdisk.h>
#include <fs/fs_subr.h>
#include <sys/modctl.h>
#include <sys/vol.h>
#include <sys/dkio.h>
#include <sys/open.h>
#include <sys/stat.h>
#include <sys/mode.h>
#include <ext2fs.h>

#define	log2(n)	ffz(~(n))

struct kmem_cache *ext2fs_inode_cache;

unsigned long
ffz(unsigned long word)
{
	unsigned long result = 0;

	while(word & 1) {
		result++;
		word >>= 1;
	}
	return result;
}

int
ext2fs_check_descriptors(struct ext2fs *fsp)
{
	int	i;
	int	desc_block = 0;
	size_t	block = fsp->e_sb.s_first_data_block;
	struct ext2_group_desc *gdp = NULL, gd;

	for (i = 0; i < fsp->e_sbi.s_groups_count; i++) {
		if ((i % EXT2_DESC_PER_BLOCK(fsp)) == 0)
			gdp = (struct ext2_group_desc *)
			    fsp->e_sbi.s_group_desc[desc_block++]->b_un.b_addr;
		gd.bg_block_bitmap = ext2fs_swap32(fsp, gdp->bg_block_bitmap);
		gd.bg_inode_bitmap = ext2fs_swap32(fsp, gdp->bg_inode_bitmap);
		gd.bg_inode_table = ext2fs_swap32(fsp, gdp->bg_inode_table);
		gd.bg_free_blocks_count =
		    ext2fs_swap16(fsp, gdp->bg_free_blocks_count);
		gd.bg_free_inodes_count =
		    ext2fs_swap16(fsp, gdp->bg_free_inodes_count);
		gd.bg_used_dirs_count =
		    ext2fs_swap16(fsp, gdp->bg_used_dirs_count);
		if ((gd.bg_block_bitmap < block) ||
		    (gd.bg_block_bitmap >= (block +
		    EXT2_BLOCKS_PER_GROUP(fsp)))) {
			cmn_err(CE_NOTE, "ext2fs: Block bitmap for group %d"
			    " not in group (block %u)!", i,
			    (unsigned long)gd.bg_block_bitmap);
			return (0);
		}
		if ((gd.bg_inode_bitmap < block) ||
		    (gd.bg_inode_bitmap >= (block +
		    EXT2_BLOCKS_PER_GROUP(fsp)))) {
			cmn_err(CE_NOTE, "ext2fs: Inode bitmap for group %d"
			    " not in group (block %u)!", i,
			    (unsigned long)gd.bg_inode_bitmap);
			return (0);
		}
		if ((gd.bg_inode_table < block) ||
		    ((gd.bg_inode_table + fsp->e_sbi.s_itb_per_group) >=
		    (block + EXT2_BLOCKS_PER_GROUP(fsp)))) {
			cmn_err(CE_NOTE, "ext2fs: Inode table for group %d"
			    " not in group (block %u)!", i,
			    (unsigned long)gd.bg_inode_table);
			return (0);
		}
		block += EXT2_BLOCKS_PER_GROUP(fsp);
		gdp++;
	}
	return (1);
}

/*
 * ufs just saves the super block in another buffer. That might be cool,
 * since then it's ready to go out via I/O.
 */
int
ext2fs_read_super(struct ext2fs *fsp)
{
	struct buf *tp, *bp = NULL;
	int	error = 0;
	int	sb_block = EXT2_SUPER_BLOCK_SIZE / DEV_BSIZE ;
	struct ext2_super_block *tsb;
	size_t	db_count;
	int	i;
	struct inode *rip;
	struct vnode *rvp = NULL;

	tp = BREAD(fsp->e_devvp->v_rdev, sb_block,
	     EXT2_SUPER_BLOCK_SIZE);
	if (tp->b_flags & B_ERROR) {
		error = tp->b_error;
		goto failed;
	} 
	tsb = (struct ext2_super_block *)tp->b_un.b_addr;
	if (tsb->s_magic != EXT2_SUPER_MAGIC) {
		/* other-endian ext2 filesystem? */
		fsp->e_flags |= EXT2FS_BYTE_SWAP;
		if (tsb->s_magic != ext2fs_swap16(fsp, EXT2_SUPER_MAGIC)) {
			fsp->e_flags &= ~EXT2FS_BYTE_SWAP;
			error = EINVAL;
			goto failed;
		}
	}
	fsp->e_sb.s_magic = ext2fs_swap16(fsp, tsb->s_magic);
	fsp->e_sb.s_inodes_count = ext2fs_swap32(fsp, tsb->s_inodes_count);
	fsp->e_sb.s_blocks_count = ext2fs_swap32(fsp, tsb->s_blocks_count);
	fsp->e_sb.s_r_blocks_count = ext2fs_swap32(fsp, tsb->s_r_blocks_count);
	fsp->e_sb.s_free_blocks_count =
	    ext2fs_swap32(fsp, tsb->s_free_blocks_count);
	fsp->e_sb.s_free_inodes_count =
	    ext2fs_swap32(fsp, tsb->s_free_inodes_count);
	fsp->e_sb.s_first_data_block =
	    ext2fs_swap32(fsp, tsb->s_first_data_block);
	fsp->e_sb.s_log_block_size = ext2fs_swap32(fsp, tsb->s_log_block_size);
	fsp->e_sb.s_log_frag_size = ext2fs_swap32(fsp, tsb->s_log_frag_size);
	fsp->e_sb.s_blocks_per_group =
	    ext2fs_swap32(fsp, tsb->s_blocks_per_group);
	fsp->e_sb.s_frags_per_group = ext2fs_swap32(fsp, tsb->s_frags_per_group);
	fsp->e_sb.s_inodes_per_group =
	    ext2fs_swap32(fsp, tsb->s_inodes_per_group);
	fsp->e_sb.s_mtime = ext2fs_swap32(fsp, tsb->s_mtime);
	fsp->e_sb.s_wtime = ext2fs_swap32(fsp, tsb->s_wtime);
	fsp->e_sb.s_mnt_count = ext2fs_swap16(fsp, tsb->s_mnt_count);
	fsp->e_sb.s_max_mnt_count = ext2fs_swap16(fsp, tsb->s_max_mnt_count);
	fsp->e_sb.s_state = ext2fs_swap16(fsp, tsb->s_state);
	fsp->e_sb.s_errors = ext2fs_swap16(fsp, tsb->s_errors);
	fsp->e_sb.s_minor_rev_level = ext2fs_swap16(fsp, tsb->s_minor_rev_level);
	fsp->e_sb.s_lastcheck = ext2fs_swap32(fsp, tsb->s_lastcheck);
	fsp->e_sb.s_checkinterval = ext2fs_swap32(fsp, tsb->s_checkinterval);
	fsp->e_sb.s_creator_os = ext2fs_swap32(fsp, tsb->s_creator_os);
	fsp->e_sb.s_rev_level = ext2fs_swap32(fsp, tsb->s_rev_level);
	fsp->e_sb.s_def_resuid = ext2fs_swap16(fsp, tsb->s_def_resuid);
	fsp->e_sb.s_def_resgid = ext2fs_swap16(fsp, tsb->s_def_resgid);
	fsp->e_sb.s_first_ino = ext2fs_swap32(fsp, tsb->s_first_ino);
	fsp->e_sb.s_inode_size = ext2fs_swap16(fsp, tsb->s_inode_size);
	fsp->e_sb.s_block_group_nr = ext2fs_swap16(fsp, tsb->s_block_group_nr);
	fsp->e_sb.s_feature_compat = ext2fs_swap32(fsp, tsb->s_feature_compat);
	fsp->e_sb.s_feature_incompat =
	    ext2fs_swap32(fsp, tsb->s_feature_incompat);
	fsp->e_sb.s_feature_ro_compat =
	    ext2fs_swap32(fsp, tsb->s_feature_ro_compat);

	tsb = &fsp->e_sb;

	if (tsb->s_rev_level > EXT2_GOOD_OLD_REV) {
		if (tsb->s_feature_incompat & ~EXT2_FEATURE_INCOMPAT_SUPP) {
			cmn_err(CE_NOTE, "ext2fs: couldn't mount because of "
			    "unsupported optional features.");
			error = EINVAL;
			goto failed;
		}
		if (!(fsp->e_vfs->vfs_flag & VFS_RDONLY) &&
		    (tsb->s_feature_ro_compat & ~EXT2_FEATURE_RO_COMPAT_SUPP)) {
			cmn_err(CE_NOTE, "ext2fs: couldn't mount rw because of "
			    "unsupported optional features.");
			error = EINVAL;
			goto failed;
		}
	}

	fsp->e_sbi.s_blocksize_bits = tsb->s_log_block_size + 10;
	fsp->e_sbi.s_blocksize = 1 << fsp->e_sbi.s_blocksize_bits;

	if ((fsp->e_sbi.s_blocksize != EXT2_SUPER_BLOCK_SIZE) &&
	    ((fsp->e_sbi.s_blocksize == 1024) ||
	    (fsp->e_sbi.s_blocksize == 2048) ||
	    (fsp->e_sbi.s_blocksize == 4096))) {
		tp->b_flags |= B_STALE | B_AGE;
		brelse(tp);
		tp = BREAD(fsp->e_devvp->v_rdev, sb_block,
		     fsp->e_sbi.s_blocksize);
		if (tp->b_flags & B_ERROR) {
			error = tp->b_error;
			goto failed;
		}
	}

	if (tsb->s_rev_level == EXT2_GOOD_OLD_REV) {
		fsp->e_sbi.s_inode_size = EXT2_GOOD_OLD_INODE_SIZE;
		fsp->e_sbi.s_first_ino = EXT2_GOOD_OLD_FIRST_INO;
	} else {
		fsp->e_sbi.s_inode_size = tsb->s_inode_size;
		fsp->e_sbi.s_first_ino = tsb->s_first_ino;
		if (fsp->e_sbi.s_inode_size != EXT2_GOOD_OLD_INODE_SIZE) {
			cmn_err(CE_NOTE, "ext2fs: unsupported inode size %d",
			    tsb->s_inode_size);
			error = EINVAL;
			goto failed;
		}
	}

	fsp->e_sbi.s_frag_size = EXT2_MIN_FRAG_SIZE << tsb->s_log_frag_size;
	if (fsp->e_sbi.s_frag_size == 0) {
		cmn_err(CE_NOTE, "ext2fs: zero frag size");
		error = EINVAL;
		goto failed;
	}
	fsp->e_sbi.s_frags_per_block = fsp->e_sbi.s_blocksize /
	    fsp->e_sbi.s_frag_size;
	fsp->e_sbi.s_blocks_per_group = tsb->s_blocks_per_group;
	fsp->e_sbi.s_frags_per_group = tsb->s_frags_per_group;
	fsp->e_sbi.s_inodes_per_group = tsb->s_inodes_per_group;
	fsp->e_sbi.s_inodes_per_block = fsp->e_sbi.s_blocksize /
	    EXT2_INODE_SIZE(fsp);
	fsp->e_sbi.s_itb_per_group = fsp->e_sbi.s_inodes_per_group /
	    fsp->e_sbi.s_inodes_per_block;
	fsp->e_sbi.s_desc_per_block = fsp->e_sbi.s_blocksize /
	    sizeof (struct ext2_group_desc);

	fsp->e_sbi.s_resuid = tsb->s_def_resuid;
	fsp->e_sbi.s_resgid = tsb->s_def_resgid;
	fsp->e_sbi.s_mount_state = tsb->s_state;

	fsp->e_sbi.s_addr_per_block_bits = log2(EXT2_ADDR_PER_BLOCK(fsp));
	fsp->e_sbi.s_desc_per_block_bits = log2(EXT2_DESC_PER_BLOCK(fsp));
	
	if (fsp->e_sbi.s_blocksize != fsp->e_sbi.s_frag_size) {
		cmn_err(CE_NOTE, "ext2fs: fragsize %u != blocksize %u "
		    "(not supported)", fsp->e_sbi.s_frag_size,
		    fsp->e_sbi.s_blocksize);
		error = EINVAL;
		goto failed;
	}

	if (fsp->e_sbi.s_blocks_per_group > (fsp->e_sbi.s_blocksize * 8)) {
		cmn_err(CE_NOTE, "ext2fs: #blocks per group too big %u",
		    fsp->e_sbi.s_blocks_per_group);
		error = EINVAL;
		goto failed;
	}

	if (fsp->e_sbi.s_frags_per_group > (fsp->e_sbi.s_blocksize * 8)) {
		cmn_err(CE_NOTE, "ext2fs: #frags per group too big %u",
		    fsp->e_sbi.s_frags_per_group);
		error = EINVAL;
		goto failed;
	}

	if (fsp->e_sbi.s_inodes_per_group > (fsp->e_sbi.s_blocksize * 8)) {
		cmn_err(CE_NOTE, "ext2fs: #inodes per group too big %u",
		    fsp->e_sbi.s_inodes_per_group);
		error = EINVAL;
		goto failed;
	}

	/* we think it is valid */
	bp = ngeteblk(fsp->e_sbi.s_blocksize);
	fsp->e_bufp = bp;
	bp->b_edev = fsp->e_devvp->v_rdev;
	bp->b_dev = cmpdev(bp->b_edev);
	bp->b_blkno = sb_block;
	bp->b_bcount = EXT2_SUPER_BLOCK_SIZE;
	bcopy(tp->b_un.b_addr, bp->b_un.b_addr, fsp->e_sbi.s_blocksize);
	tp->b_flags |= B_STALE | B_AGE;
	brelse(tp);
	tp = NULL;

	fsp->e_sbi.s_groups_count = (tsb->s_blocks_count -
	    tsb->s_first_data_block + EXT2_BLOCKS_PER_GROUP(fsp) - 1) /
	    EXT2_BLOCKS_PER_GROUP(fsp);

	db_count = (fsp->e_sbi.s_groups_count + EXT2_DESC_PER_BLOCK(fsp) - 1) /
	    EXT2_DESC_PER_BLOCK(fsp);
	fsp->e_sbi.s_group_desc = kmem_zalloc(db_count * sizeof (struct buf *),
	    KM_SLEEP);
	
	for (i = 0; i < db_count; i++) {
		int blkno;

		blkno = sb_block + lblkno(fsp, i + 1);
		fsp->e_sbi.s_group_desc[i] = BREAD(fsp->e_devvp->v_rdev, blkno,
		     fsp->e_sbi.s_blocksize);
		if (fsp->e_sbi.s_group_desc[i] == NULL) {
			cmn_err(CE_NOTE, "ext2fs: unable to read group "
			    "descriptors");
			goto failed;
		}
	}

	if (!ext2fs_check_descriptors(fsp)) {
		cmn_err(CE_NOTE, "ext2fs: group descriptors corrupted!");
		goto failed;
	}

	for (i = 0; i < EXT2_MAX_GROUP_LOADED; i++) {
		fsp->e_sbi.s_inode_bitmap_number[i] = 0;
		fsp->e_sbi.s_inode_bitmap[i] = NULL;
		fsp->e_sbi.s_block_bitmap_number[i] = 0;
		fsp->e_sbi.s_block_bitmap[i] = NULL;
	}
	fsp->e_sbi.s_loaded_inode_bitmaps = 0;
	fsp->e_sbi.s_loaded_block_bitmaps = 0;
	fsp->e_sbi.s_db_per_group = db_count;

	error = ext2fs_iget(fsp->e_vfs, EXT2_ROOT_INO, &rip, kcred);
	if (error)
		goto failed;
	
	fsp->e_root = ITOV(rip);
	/*
	if ((rip->i_mode & IFMT) != IFDIR) {
		error = EIO;
		goto out;
	}
	*/
	rvp = fsp->e_root;

	mutex_enter(&rvp->v_lock);
	rvp->v_flag |= VROOT;
	mutex_exit(&rvp->v_lock);

	return (0);
failed:
	if (tp) {
		tp->b_flags |= B_STALE | B_AGE;
		brelse(tp);
	}
	if (bp) {
		bp->b_flags |= B_STALE | B_AGE;
		brelse(bp);
	}
	if (error == 0)
		error = EIO;
	if (fsp->e_sbi.s_group_desc) {
		for (i = 0; i < db_count; i++) {
			bp = fsp->e_sbi.s_group_desc[i];
			fsp->e_sbi.s_group_desc[i] = NULL;
			if (bp) {
				bp->b_flags |= B_STALE | B_AGE;
				brelse(bp);
			}
		}
		kmem_free(fsp->e_sbi.s_group_desc,
		    db_count * sizeof (struct buf *));
		fsp->e_sbi.s_group_desc = NULL;
	}
	return (error);
}

void
ext2fs_free_super(struct ext2fs *fsp)
{
	int	i;
	struct buf *bp;
	int	db_count;

	if (fsp->e_sbi.s_group_desc) {
		db_count = fsp->e_sbi.s_db_per_group;

		for (i = 0; i < db_count; i++) {
			bp = fsp->e_sbi.s_group_desc[i];
			if (bp) {
				fsp->e_sbi.s_group_desc[i] = NULL;
				bp->b_flags |= B_STALE | B_AGE;
				brelse(bp);
			}
		}
		kmem_free(fsp->e_sbi.s_group_desc,
		    db_count * sizeof (struct buf *));
		fsp->e_sbi.s_group_desc = NULL;
	}
	if (fsp->e_bufp) {
		fsp->e_bufp->b_flags |= B_STALE | B_AGE;
		brelse(fsp->e_bufp);
		fsp->e_bufp = NULL;
	}
}

void
ext2fs_read_inode(struct inode *inode)
{
	struct buf *bp;
	struct ext2_inode *raw_inode;
	size_t block_group;
	size_t group_desc;
	size_t desc;
	size_t block;
	size_t offset;
	struct ext2_group_desc *gdp;
	uint32_t bg_inode_table;
	struct ext2fs *fsp = inode->i_fs;

	if ((inode->i_number != EXT2_ROOT_INO &&
	    inode->i_number != EXT2_ACL_IDX_INO &&
	    inode->i_number != EXT2_ACL_DATA_INO &&
	    inode->i_number < EXT2_FIRST_INO(fsp)) ||
	    inode->i_number > fsp->e_sb.s_inodes_count) {
		cmn_err(CE_PANIC, "ext2fs_read_inode: bad inode number: %lu",
		    inode->i_number);
	}
	block_group = (inode->i_number - 1) / EXT2_INODES_PER_GROUP(fsp);
	if (block_group >= fsp->e_sbi.s_groups_count) {
		cmn_err(CE_PANIC, "ext2fs_read_inode: group >= groups count");
	}
	group_desc = block_group >> EXT2_DESC_PER_BLOCK_BITS(fsp);
	desc = block_group & (EXT2_DESC_PER_BLOCK(fsp) - 1);
	bp = fsp->e_sbi.s_group_desc[group_desc];
	if (!bp) {
		cmn_err(CE_PANIC, "ext2fs_read_inode: Descriptor not loaded");
	}
	
	gdp = (struct ext2_group_desc *)bp->b_un.b_addr;
	/*
	 * Figure out the offset within the block group inode table
	 */
	offset = ((inode->i_number - 1) % EXT2_INODES_PER_GROUP(fsp)) *
		EXT2_INODE_SIZE(fsp);
	bg_inode_table = ext2fs_swap32(fsp, gdp[desc].bg_inode_table);
	block = bg_inode_table + (offset >> EXT2_BLOCK_SIZE_BITS(fsp));
	if (!(bp = BREAD(inode->i_dev, lblkno(fsp, block),
	    EXT2_BLOCK_SIZE(fsp)))) {
		cmn_err(CE_PANIC,
		    "ext2fs_read_inode: unable to read inode block - "
		    "inode=%lu, block=%lu", inode->i_number, block);
	}
	
	offset &= (EXT2_BLOCK_SIZE(fsp) - 1);
	raw_inode = (struct ext2_inode *)(bp->b_un.b_addr + offset);

	inode->i_mode = ext2fs_swap16(fsp, raw_inode->i_mode);
	inode->i_uid = ext2fs_swap16(fsp, raw_inode->i_uid);
	inode->i_size = ext2fs_swap32(fsp, raw_inode->i_size);
	inode->i_atime = ext2fs_swap32(fsp, raw_inode->i_atime);
	inode->i_ctime = ext2fs_swap32(fsp, raw_inode->i_ctime);
	inode->i_mtime = ext2fs_swap32(fsp, raw_inode->i_mtime);
	inode->i_dtime = ext2fs_swap32(fsp, raw_inode->i_dtime);
	inode->i_gid = ext2fs_swap16(fsp, raw_inode->i_gid);
	inode->i_links_count = ext2fs_swap16(fsp, raw_inode->i_links_count);
	inode->i_blocks = ext2fs_swap32(fsp, raw_inode->i_blocks);
	inode->i_flags = ext2fs_swap32(fsp, raw_inode->i_flags);
	inode->i_reserved1 = ext2fs_swap32(fsp, raw_inode->i_reserved1);
	for (block = 0; block < EXT2_N_BLOCKS; block++) {
		inode->i_data[block] =
		    ext2fs_swap32(fsp, raw_inode->i_block[block]);
	}
	inode->i_version = ext2fs_swap32(fsp, raw_inode->i_version);
	inode->i_file_acl = ext2fs_swap32(fsp, raw_inode->i_file_acl);
	inode->i_dir_acl = ext2fs_swap32(fsp, raw_inode->i_dir_acl);
	inode->i_faddr = ext2fs_swap32(fsp, raw_inode->i_faddr);
	inode->i_frag = raw_inode->i_frag;
	inode->i_fsize = raw_inode->i_fsize;
	inode->i_pad1 = ext2fs_swap16(fsp, raw_inode->i_pad1);
	inode->i_reserved2[0] = ext2fs_swap32(fsp, raw_inode->i_reserved2[0]);
	inode->i_reserved2[1] = ext2fs_swap32(fsp, raw_inode->i_reserved2[1]);

	inode->i_new_inode = 0;
	inode->i_osync = 0;
	inode->i_block_group = block_group;
	inode->i_next_alloc_block = 0;
	inode->i_next_alloc_goal = 0;
	brelse(bp);
}

void
ext2fs_free_inode(struct inode *ip)
{
	kmem_cache_free(ext2fs_inode_cache, ip);
}

struct inode *
ext2fs_alloc_inode(struct ext2fs *fsp, ino_t ino)
{
	struct inode *ip;
	vnode_t	*vp;

	ip = kmem_cache_alloc(ext2fs_inode_cache, KM_SLEEP);
	(void) bzero(ip, sizeof (struct inode));

	ip->i_fs = fsp;
	ip->i_new_inode = 1;
	ip->i_number = ino;
	ip->i_dev = fsp->e_xdev;

	ext2fs_read_inode(ip);

	/*
	 * Initialize vnode
	 */
	vp = ITOV(ip);
	vp->v_flag = 0;
	vp->v_count = 1;
	vp->v_op = &ext2fs_vnodeops;
	vp->v_vfsp = fsp->e_vfs;
	vp->v_data = (caddr_t)ip;
	vp->v_vfsmountedhere = NULL;
	vp->v_stream = NULL;
	vp->v_pages = NULL;
	vp->v_type = IFTOVT((mode_t)ip->i_mode);
	vp->v_rdev = 0;
	vp->v_filocks = NULL;
	vp->v_shrlocks = NULL;

	return (ip);
}

int
ext2fs_iget(struct vfs *vfsp, ino_t ino, struct inode **ipp, struct cred *cr)
{
	struct ext2fs *fsp = (struct ext2fs *)vfsp->vfs_data;
	struct inode	*ip, *sp;
	union ihead	*ih;
	kmutex_t	*ihm;
	int	hno;
	dev_t	vfs_dev;
	struct vnode *vp;

	/*
	 * Lookup inode in cache.
	 */
	vfs_dev = vfsp->vfs_dev;
	hno = INOHASH(vfs_dev, ino);
	ih = &ihead[hno];
	ihm = &ih_lock[hno];

again:
	mutex_enter(ihm);
	for (ip = ih->ih_chain[0]; ip != (struct inode *)ih; ip = ip->i_forw) {
		if (ino != ip->i_number || vfs_dev != ip->i_dev)
			continue;

		/*
		 * Found the interesting inode; hold it and drop the cache lock
		 */
		vp = ITOV(ip);  /* for locknest */
		VN_HOLD(vp);

		*ipp = ip;
		mutex_exit(ihm);
		return (0);
	}
	mutex_exit(ihm);

	/* inode was not in cache */

	/* construct an in-memory inode and go read it */
	ip = ext2fs_alloc_inode(fsp, ino);

	/* insert it into cache */
	mutex_enter(ihm);
	insque(ip, ih);
	mutex_exit(ihm);

	*ipp = ip;
	return (0);
}

static int
ext2_alloc_block(struct inode *inode, unsigned long goal, int *err)
{
	int	result;

	result = ext2_new_block (inode, goal, 0, 0, err);
	return (result);
}

static struct buf *
inode_getblk(struct inode *inode, int nr, int create, int new_block, int *err)
{
	uint32_t *p;
	int tmp, goal = 0;
	struct buf *result;
	struct ext2fs *fsp = inode->i_fs;
	int blocks = EXT2_BLOCK_SIZE(fsp) / DEV_BSIZE;

	p = inode->i_data + nr;
repeat:
	tmp = *p;
	if (tmp) {
		result = getblk(inode->i_dev, lblkno(fsp, tmp),
		    EXT2_BLOCK_SIZE(fsp));
		if (tmp == *p)
			return result;
		brelse(result);
		goto repeat;
	}
	if (!create || new_block >= 
	    (P_CURLIMIT(curproc, RLIMIT_FSIZE) >> EXT2_BLOCK_SIZE_BITS(fsp))) {
		*err = EFBIG;
		return NULL;
	}
	if (inode->i_next_alloc_block == new_block)
		goal = inode->i_next_alloc_goal;

	if (!goal) {
		for (tmp = nr - 1; tmp >= 0; tmp--) {
			if (inode->i_data[tmp]) {
				goal = inode->i_data[tmp];
				break;
			}
		}
		if (!goal)
			goal = (inode->i_block_group *
			    EXT2_BLOCKS_PER_GROUP(fsp)) +
			    fsp->e_sb.s_first_data_block;
	}

	tmp = ext2_alloc_block(inode, goal, err);
	if (!tmp)
		return NULL;
	result = getblk(inode->i_dev, lblkno(fsp, tmp), EXT2_BLOCK_SIZE(fsp));
	if (*p) {
		ext2_free_blocks(inode, tmp, 1);
		brelse(result);
		goto repeat;
	}
	*p = tmp;
	inode->i_next_alloc_block = new_block;
	inode->i_next_alloc_goal = tmp;
	/*
	inode->i_ctime = CURRENT_TIME;
	 */
	inode->i_blocks += blocks;
/*
	if (IS_SYNC(inode) || inode->u.ext2_i.i_osync)
		ext2_sync_inode (inode);
	else
		inode->i_dirt = 1;
*/
	return result;
}

static struct buf *
block_getblk(struct inode *inode, struct buf *bh, int nr, int create,
    int blocksize, int new_block, int *err)
{
	int tmp, goal = 0;
	uint32_t *p;
	struct buf *result;
	struct ext2fs *fsp = inode->i_fs;
	int blocks = EXT2_BLOCK_SIZE(fsp) / DEV_BSIZE;

	if (!bh)
		return NULL;
	/* XXX: go read the block */
	if (!(bh->b_flags & B_DONE)) {
		bh->b_flags |= B_READ;
		(void) bdev_strategy(bh);
		biowait(bh);
	}
	p = (uint32_t *)bh->b_un.b_addr + nr;
repeat:
	tmp = ext2fs_swap32(fsp, *p);
	if (tmp) {
		result = getblk(bh->b_edev, lblkno(fsp, tmp), blocksize);
		if (tmp == ext2fs_swap32(fsp, *p)) {
			brelse(bh);
			return result;
		}
		brelse(result);
		goto repeat;
	}
	if (!create || new_block >= 
	    (P_CURLIMIT(curproc, RLIMIT_FSIZE) >> EXT2_BLOCK_SIZE_BITS(fsp))) {
		*err = EFBIG;
		brelse(bh);
		return NULL;
	}
	if (inode->i_next_alloc_block == new_block)
		goal = inode->i_next_alloc_goal;
	if (!goal) {
		for (tmp = nr - 1; tmp >= 0; tmp--) {
			if (((uint32_t *)bh->b_un.b_addr)[tmp]) {
				goal = ext2fs_swap32(fsp,
				    ((uint32_t *)bh->b_un.b_addr)[tmp]);
				break;
			}
		}
		if (!goal)
			goal = rblkno(fsp, bh->b_blkno);
	}
	tmp = ext2_alloc_block(inode, goal, err);
	if (!tmp) {
		brelse(bh);
		return NULL;
	}
	result = getblk(bh->b_edev, lblkno(fsp, tmp), blocksize);
	if (*p) {
		ext2_free_blocks(inode, tmp, 1);
		brelse(result);
		goto repeat;
	}
	*p = ext2fs_swap32(fsp, tmp);
/*
	mark_buffer_dirty(bh, 1);
	if (IS_SYNC(inode) || inode->u.ext2_i.i_osync) {
		ll_rw_block (WRITE, 1, &bh);
		wait_on_buffer (bh);
	}
	inode->i_ctime = CURRENT_TIME;
	inode->i_dirt = 1;
*/
	inode->i_blocks += blocks;
	inode->i_next_alloc_block = new_block;
	inode->i_next_alloc_goal = tmp;
	brelse(bh);
	return result;
}

struct buf *
ext2_getblk(struct inode *inode, long block, int create, int *err)
{
	struct buf *bh;
	unsigned long b;
	struct ext2fs *fsp = inode->i_fs;
	unsigned long addr_per_block = EXT2_ADDR_PER_BLOCK(fsp);
	int addr_per_block_bits = EXT2_ADDR_PER_BLOCK_BITS(fsp);

	*err = EIO;
	if (block < 0) {
		return NULL;
	}
	if (block > EXT2_NDIR_BLOCKS + addr_per_block +
		(1 << (addr_per_block_bits * 2)) +
		((1 << (addr_per_block_bits * 2)) << addr_per_block_bits)) {
		return NULL;
	}
	/*
	 * If this is a sequential block allocation, set the next_alloc_block
	 * to this block now so that all the indblock and data block
	 * allocations use the same goal zone
	 */

	if (block == (inode->i_next_alloc_block + 1)) {
		inode->i_next_alloc_block++;
		inode->i_next_alloc_goal++;
	}

	*err = ENOSPC;
	b = block;
	if (block < EXT2_NDIR_BLOCKS)
		return inode_getblk(inode, block, create, b, err);
	block -= EXT2_NDIR_BLOCKS;
	if (block < addr_per_block) {
		bh = inode_getblk(inode, EXT2_IND_BLOCK, create, b, err);
		return (block_getblk(inode, bh, block, create,
		    EXT2_BLOCK_SIZE(fsp), b, err));
	}
	block -= addr_per_block;
	if (block < (1 << (addr_per_block_bits * 2))) {
		bh = inode_getblk(inode, EXT2_DIND_BLOCK, create, b, err);
		bh = block_getblk(inode, bh, block >> addr_per_block_bits,
		    create, EXT2_BLOCK_SIZE(fsp), b, err);
		return (block_getblk(inode, bh, block & (addr_per_block - 1),
		    create, EXT2_BLOCK_SIZE(fsp), b, err));
	}
	block -= (1 << (addr_per_block_bits * 2));
	bh = inode_getblk(inode, EXT2_TIND_BLOCK, create, b, err);
	bh = block_getblk(inode, bh, block >> (addr_per_block_bits * 2),
	    create, EXT2_BLOCK_SIZE(fsp), b, err);
	bh = block_getblk(inode, bh,
	    (block >> addr_per_block_bits) & (addr_per_block - 1),
	    create, EXT2_BLOCK_SIZE(fsp), b, err);
	return (block_getblk(inode, bh, block & (addr_per_block - 1), create,
	    EXT2_BLOCK_SIZE(fsp), b, err));
}

struct buf *
ext2fs_bread(struct inode *inode, int block, int create, int *err)
{
	struct buf *bh;

	bh = ext2_getblk(inode, block, create, err);
	if (!bh) {
		return bh;
	}
	if (!(bh->b_flags & B_DONE)) {
		bh->b_flags |= B_READ;
		(void) bdev_strategy(bh);
		biowait(bh);
	}
	*err = geterror(bh);
	if (*err != 0) {
		brelse (bh);
		return (NULL);
	}
	return (bh);
}

int
ext2_new_block(const struct inode *inode, unsigned long goal,
   uint32_t *prealloc_count, uint32_t *prealloc_block, int *err)
{
	*err = EROFS;
	return (0);
}

void
ext2_free_blocks (const struct inode * inode, unsigned long block,
    unsigned long count)
{
}

/*
 * Check mode permission on inode.  Mode is READ, WRITE or EXEC.
 * In the case of WRITE, the read-only status of the file system
 * is checked.  The mode is shifted to select the owner/group/other
 * fields.  The super user is granted all permissions except
 * writing to read-only file systems.
 */
int
ext2fs_iaccess(struct inode *ip, int mode, struct cred *cr)
{
	if (ip->i_fs == NULL)
		return (EIO);
	if (mode & IWRITE) {
		/*
		 * Disallow write attempts on read-only
		 * file systems, unless the file is a block
		 * or character device or a FIFO.
		 */
		return (EROFS);
#ifdef notdef
		if (ip->i_fs->fs_ronly != 0) {
			if ((ip->i_mode & IFMT) != IFCHR &&
			    (ip->i_mode & IFMT) != IFBLK &&
			    (ip->i_mode & IFMT) != IFIFO) {
				return (EROFS);
			}
		}
#endif
	}
	/*
	 * If you're the super-user,
	 * you always get access.
	 */
	if (cr->cr_uid == 0)
		return (0);
	/*
	 * Access check is based on only
	 * one of owner, group, public.
	 * If not owner, then check group.
	 * If not a member of the group, then
	 * check public access.
	 */
	if (cr->cr_uid != ip->i_uid) {
		mode >>= 3;
		if (!groupmember((uid_t)ip->i_gid, cr))
			mode >>= 3;
	}
	if ((ip->i_mode & mode) == mode)
		return (0);
	return (EACCES);
}

uint16_t
ext2fs_swap16(struct ext2fs *fsp, uint16_t value)
{
	if (!(fsp->e_flags & EXT2FS_BYTE_SWAP))
		return (value);
	return ((value >> 8) | (value << 8));
}

uint32_t
ext2fs_swap32(struct ext2fs *fsp, uint32_t value)
{
	if (!(fsp->e_flags & EXT2FS_BYTE_SWAP))
		return (value);
	return ((value >> 24) | ((value >> 8) & 0xff00) |
	    ((value << 8) & 0xff0000) | (value << 24));
}
